/** This class is for representing complex lines.
 * A line is constructed by way of a set of blocks 
 * @see block2D.
 *  \author Alejandro Pauly, Mark McKenney
 *  \version 0.1
 *  \date    2005-10-05
 *  \version 0.2
 *  \date    2005-10-14 
 *  \version 0.4 Removed Impl stuff, fixing for mSlob
 *  \date    2006-10-13 
 */

#ifndef LINE2D_H
#define LINE2D_H

#include "constants.h"
#include "num.h"
#include "hseg2D.h"
#include "mSlob.h"
#include "block2D.h"
#include "point2D.h"
//#include "region2D.h"
#include "invalidObjectException.h"

#include "vectorBinarySearchWrapper.h"

#include "statusHseg2D.h"
#include "EventSchedule.h"
#include "SweepLine.h"
#include "dvector.h"
#include "delaunay.h"
#include "convexHull.h"


// #include "sweepLineSingleStatus.h"
// #include "MEventSchedule.h"

namespace stu{
	using namespace stu;

	class block2D;
//	class point2D;
	class region2D;
	class CDelaunay;
	class CConvexHull;
	
	class line2D{
	
		public:
			friend class block2D;
						
			class Iterator{ //iterates over each poi2D in list
				long int current;
				line2D* parent;
				
				public:
					Iterator();
					Iterator(line2D* par);
					hseg2D* get();
					bool begin();
					bool next();
					bool prev();
					bool end(); 
					Iterator operator+( int num );
					Iterator operator-( int num );
					Iterator operator++( int dummy );
					Iterator operator--( int dummy );
					Iterator& operator+=( int num );
					Iterator& operator-=( int num );
			};
			
			line2D(bool create=true);
			line2D(const Locator& m);
			line2D(const Locator& m, const block2D& b);
			line2D(const Locator& m, const block2D bs[], uint length);
			line2D(const Locator& m, const line2D &ls);
			
			~line2D();
				
			hseg2D operator[]( uint index );
			uint size() const; //the number of half segments				
			
			//constructor ops
		
			block2D addBlock(const block2D &b); //returns version of block stored as subobject of this line
			
 			block2D addBlock(); //adds an empty block, and returns it for editing
				
			//adds all blocks, leaves checking to finalize
			void addBlocks(const block2D bs[], uint length) throw (invalidObjectException); 
		
			//removes a block and returns a copy of the removed block in up, to allow for updates
			bool removeBlock(uint idx);
				
			block2D getBlock(uint idx); //returns the idx block of this line
			
			bool validate();
				
			uint getSegmentCount() const; //returns the number of segments for the whole line
			
			uint getBlockCount() const; //number of blocks
				
			rat & getLength() const; //returns the perimeter. sum of lenghts of segments
			
			rat  distance(const line2D &other) const; //returns the minimum distance between this and other
			rat  distance(const point2D &other) const; //returns the minimum distance between this and other
				
			bool isEmpty() const; //returns getSize()==0
			bool isValid() const; 
			mbb2D& getmbb() const;
		
			Iterator getIterator() const;
				
			//homogeneous ops
			line2D & intersection( line2D &other);
			line2D & sunion( line2D &other);
			line2D & difference( line2D &other);
		
			point2D & common_points( line2D &other);
			
/*				//heterogeneous ops
			line2D & intersection(const region2D &other);
			line2D & difference(const region2D &other);*/
				
/*			rat & distance(const region2D &other);*/
			
			//Output operator
			friend std::ostream &operator << (std::ostream &ostr, const line2D &rhs) throw( undefinedSpatialObjectException );
			
			/* compute convex Hull of all the points in the lines
			   using Granham's algorithm complexity O(nlogn)
			*/
			region2D convexHull();
			/* compute delaunay triangulation of all the points in the lines
			   using incremental algorithm with an average complexity of O(nlogn)
			*/
			line2D delaunayTriangulation();

			/* compute voronoi diagram */	
			line2D voronoi();	

		private:
			Locator global; //should have two substructures, the first with hseg seq, the second with blocks
			Locator hsegs;
			Locator blks;
			
			static short int FLAGS; // num of flags before hsegs start
			
			bool valid; //valid is stored as the first subobject of hsegs
			bool changed; //keeps track if object was changed to reset valid flag
			bool inBatch; //to avoid writing valid flag more than necessary
			
			//returns locator to newly added hseg
			Locator addhseg(const hseg2D &h) throw (invalidObjectException);
			void removeSegment(const seg2D &s)  throw (undefinedSpatialObjectException);
			uint removeHalfSegment(const hseg2D &s)  throw (runtime_error, invalidObjectException);
						
			void saveValid();
			void loadValid();		
				
			void breakIntoBlocks( block2D &allResultSegments, line2D &resultLine );

// 			uint getBlockSegmentCount(uint bid);
			
/*			std::vector<hseg2D>& findConnectedHsegs(const hseg2D& current, std::vector<hseg2D> & nothere, ulong bid);
			uint connectedFromHere(const hseg2D& origin, std::vector<hseg2D> & nothere, ulong bid);
			poi2D& findFirstHalfSeg(ulong bid);
			void checkIntersections( hseg2D &H1, hseg2D &H2, std::vector<hseg2D> &vec,
									 MEventSchedule<hseg2D, hseg2D> EE, sweepLineSingleStatus<hseg2D> &SLSS );
			std::vector<hseg2D>::iterator findTwin( hseg2D &sibling);*/
			/*
			get dominating points in lines for convex hull and delaunay triangulation algorithm
			*/
			void getAllDominates(vector<poi2D>& m_poi_set);
	};

}
#endif

