/** This class is used for representing blocks of region2D objects.
 * A block represents a single connected component of a region2D object.
 *	@see region2D
 *  \author Alejandro Pauly, Mark McKenney
 *  \version 0.1
 *  \date    2005-10-05
 *  \version 0.2
 *  \date    2005-10-14
 *  \version 0.3
 *  \date    2005-11-25 
 *  \version 0.4 Removed Impl stuff, fixing for mSlob
 *  \date    2006-10-13 
 */

#ifndef POLY2D_H
#define POLY2D_H

#include "hseg2D.h"
#include "seg2D.h"
#include "num.h"
#include "mSlob.h"
#include "face2D.h"
#include "region2D.h"
#include "invalidObjectException.h"
#include "unrepairableObjectException.h"
#include "vectorBinarySearchWrapper.h"

#include "statusHseg2D.h"
#include "EventSchedule.h"
#include "SweepLine.h"
#include "dvector.h"

//#include "sweepLineSingleStatus.h"
//#include "MEventSchedule.h"

namespace stu{

	using namespace stu;

	class region2D;
	class face2D;
	
	class poly2D{
		friend class region2D;
		friend class face2D;
		
		public:
			class Iterator{ //iterates over each hseg2D in list
				long current;
				poly2D* parent;
				
				public:
					Iterator();
					Iterator(poly2D* par);
					Iterator(poly2D* par, long start);
					hseg2D* get();
					bool next();
					bool prev();
					bool end(); 
					Iterator operator+( int num );
					Iterator operator-( int num );
					Iterator operator++( int dummy );
					Iterator operator--( int dummy );
					Iterator& operator+=( int num );
					Iterator& operator-=( int num );
			};
			
			poly2D( bool isOuterCycle=true ); //default
			poly2D(const face2D& f, const Locator& p, bool isOuterCycle=true ); //region, face, poly: create empty poly for f in r or simply load if it already has something. If it loads, then it ignores isOuterCycle, otherwise, it sets it
			poly2D(const face2D& f); //create this as an empty poly of f which is a face in r
			~poly2D();
				
			hseg2D operator[]( uint index );
			uint size() const; //the number of half segments
			
			//constructor ops
			bool addSegment(const seg2D &s);
				
				//adds all segs, leaves checking to finalize

			bool addSegments(const seg2D ss[], uint length); // throw (validObjectException); 
			
			bool addhseg(const hseg2D &h)throw (runtime_error, invalidObjectException);
			
			void removeSegment(const seg2D &s)  throw (undefinedSpatialObjectException);
				
			bool validate();
	
			bool isValid() const; 
				
			uint getSegmentCount() const; //returns the number of segments for the whole block
								
			rat getLength() const; //returns the perimeter. sum of lengths of segments
				
			bool isEmpty() const; //returns getSize()==0

			bool isOuterCycle() const; //return whether this polygon represents an outer cycle in its parent face. hole if not
			
			mbb2D getmbb() const; //return mbb for this polygon

			Iterator getIterator(long start=-1) const; //sorted in half-segment order
			
			friend std::ostream& operator << (std::ostream &ostr, const poly2D &rhs) throw( invalidObjectException );
			std::ostream & debug(std::ostream &ostr) throw( undefinedSpatialObjectException );
			
		private:
			Locator global;
			Locator myface;
			Locator elements;
			//elements locates a subobject of a locator to a face located by myface (pointed by parent) which is an object in a secondary structure of the global mSlob located by global and representing the region2D pointed by grandpa below
			
			face2D *parent;
			region2D *grandpa; //pointer to its grandparent...necessary to add to primary
			
			bool isOuter; //indicates whether this polygon is an outer cycle or a hole of its parent face, by default its an outer
			bool valid;
			bool changed; //to keep whether the object was changed and so the valid flag needs to be rewritten
			bool inBatch; ////to avoid writing valid flag more than necessary

			static short int FLAGS;
			
			void saveValid();
			void loadValid();
			void saveOuter();
			void loadOuter();
			void saveFlags();
			void loadFlags();
			void invalidate(); //used to set valid=false when changes occur
			
			void removeMe(); //empties all subobjects of this poly
			long findNextsInChain(const hseg2D& cur, long curloc, dvector<hseg2D>& result, stuVector<long>& locs);
			hseg2D findBrother(const hseg2D& cur, long& loc);
			
/*			poi2D& findFirstHalfSeg();
			uint connectedFromHere(const hseg2D& origin, std::vector<hseg2D> & nothere);
			std::vector<hseg2D>& findConnectedHsegs(const hseg2D& current, std::vector<hseg2D> & nothere );
			void checkIntersections( hseg2D &H1, hseg2D &H2, std::vector<hseg2D> &vec,
			MEventSchedule<hseg2D, hseg2D> EE, sweepLineSingleStatus<hseg2D> &SLSS );
			hseg2D& findTwin( hseg2D &sibling);*/
	};
}
#endif

