/** This class is for representing complex regions.
 * A line is constructed by way of a set of faces
 * @see face2D.
 *  \author Alejandro Pauly, Mark McKenney
 *  \version 0.1
 *  \date    2005-10-05
 *  \version 0.2
 *  \date    2005-10-14 
 *  \version 0.4 Removed Impl stuff, fixing for mSlob
 *  \date    2006-10-31 
 */

#ifndef REGION2D_H
#define REGION2D_H

#include "constants.h"
#include "num.h"
#include "hseg2D.h"
#include "mSlob.h"
#include "poly2D.h"
#include "face2D.h"
#include "point2D.h"
#include "line2D.h"
#include "invalidObjectException.h"
#include "dvector.h"
#include "delaunay.h"
#include "convexHull.h"

// #include "sweepLineSingleStatus.h"
// #include "MEventSchedule.h"

namespace stu{
	using namespace stu;

	class poly2D;
	class face2D;
	class Algorithm_rr;
	class line2D;
	class poi2D;
	class CDelaunay;
	class CConvexHull;

	class region2D{
	
		public:
			friend class poly2D;
			friend class face2D;
			
			class Iterator{ //iterates over each poi2D in list
				long current;
				region2D* parent;
				
				public:
					Iterator();
					Iterator(region2D* par);
					hseg2D* get();
					bool next();
					bool prev();
					bool begin();
					bool end(); 
					Iterator operator+( int num );
					Iterator operator-( int num );
					Iterator operator++( int dummy );
					Iterator operator--( int dummy );
					Iterator& operator+=( int num );
					Iterator& operator-=( int num );
			};
			
			region2D(bool create=true);
			region2D(const Locator& m);
			region2D(const Locator& m, const face2D& f);
			region2D(const Locator& m, const face2D fs[], uint length);
			region2D(const Locator& m, const region2D &fs);
			
			~region2D();
				
			hseg2D operator[]( uint index );
			uint size() const; //the number of half segments
			
			//constructor ops
		
			face2D addFace(const face2D &b); //returns version of face stored as subobject of this region
			
 			face2D addFace(); //adds an empty face, and returns it for editing
				
			//adds all blocks, leaves checking to finalize
			void addFaces(const face2D bs[], uint length) throw (invalidObjectException); 
		
			//removes a block and returns a copy of the removed block in up, to allow for updates
			bool removeFace(uint idx);
				
			face2D getFace(uint idx); //returns the idx block of this line
			
			bool validate();
				
			uint getSegmentCount() const; //returns the number of segments for the whole line
			
			uint getFaceCount() const; //number of blocks
				
			rat getLength() const; //returns the perimeter. sum of lenghts of segments
			rat getArea() const; //returns the area of this region2D
				
			bool isEmpty() const; //returns getSegmentCount()==0
			bool isValid() const; 
			mbb2D getmbb() const;
		
			Iterator getIterator() const;
				
			//homogeneous ops
			region2D intersection(const region2D &other)  const throw (runtime_error);
			region2D sunion(const region2D &other) const;
			region2D difference(const region2D &other) const;
		
			rat  distance(const point2D &other) const; //returns the minimum distance between this and other
			rat  distance(const line2D &other) const; //returns the minimum distance between this and other
			rat  distance(const region2D &other) const; //returns the minimum distance between this and other
				
			//Output operator
			friend std::ostream &operator << (std::ostream &ostr, const region2D &rhs) throw( undefinedSpatialObjectException );
			
			/* compute convex Hull of all the points in the region
			   using Granham's algorithm complexity O(nlogn)
			*/
			region2D convexHull();
			/* compute delaunay triangulation of all the points in the region
			   using incremental algorithm with an average complexity of O(nlogn)
			*/
			line2D delaunayTriangulation();
			
			line2D voronoi();
			
			

		private:
			Locator global; //should have two substructures, the first with hseg seq, the second with blocks
			Locator hsegs;
			Locator faces;
			
			bool valid; //valid is stored as the first subobject of hsegs
			bool changed; //keeps track if object was changed to reset valid flag
			bool inBatch; //to avoid writing valid flag more than necessary

			static short int FLAGS;
			
			//returns locator to newly added hseg
			Locator addhseg(const hseg2D &h) throw (invalidObjectException);
			void removeSegment(const seg2D &s)  throw (undefinedSpatialObjectException);
			uint removeHalfSegment(const hseg2D &s)  throw (invalidObjectException); //returns idx formerly occupied by removed hseg
								
			void saveValid();
			void loadValid();	
			void invalidate(); //used to set valid=false when changes occur
// 			uint getFaceSegmentCount(uint bid);
			
/*			std::vector<hseg2D>& findConnectedHsegs(const hseg2D& current, std::vector<hseg2D> & nothere, ulong bid);
			uint connectedFromHere(const hseg2D& origin, std::vector<hseg2D> & nothere, ulong bid);
			poi2D& findFirstHalfSeg(ulong bid);
			void checkIntersections( hseg2D &H1, hseg2D &H2, std::vector<hseg2D> &vec,
									 MEventSchedule<hseg2D, hseg2D> EE, sweepLineSingleStatus<hseg2D> &SLSS );
			std::vector<hseg2D>::iterator findTwin( hseg2D &sibling);*/
			/*
			get dominating points in region for convex hull and delaunay triangulation algorithm
			*/
			void getAllDominates(vector<poi2D>& m_poi_set);	
	};

}
#endif


